/***********************************************************************/
/*                                                                     */
/*                                          O                          */
/*                             OOOOOOOOOOOOO                           */
/*                            OOOOOOOOOOOOO                            */
/*                           O  OO     OO                              */
/*                              OO     OO                              */
/*                              OO     OO                              */
/*                              OO     OO                              */
/*                             OO       OO                             */
/*                            OO         OO                            */
/*                                                                     */
/*                                                                     */
/* Program:                                                            */
/* --------                                                            */
/*                                                                     */
/* sspienr.c                                                           */
/*                                                                     */
/*                                                                     */
/* Version:                                                            */
/* --------                                                            */
/*                                                                     */
/* First version: September 1993                                       */
/* Latest update: 1998-06-22                                           */
/*                                                                     */
/*                                                                     */
/* Author of the program:                                              */
/* ----------------------                                              */
/*                                                                     */
/* Stefan Spaennare  Lund, Sweden                                      */
/*                                                                     */
/* Email: stefans@astro.lu.se                                          */
/*                                                                     */
/* If you have problems with the program please let me know !          */
/*                                                                     */
/*                                                                     */
/* Copy-right:                                                         */
/* -----------                                                         */
/*                                                                     */
/* This program is free to copy if this information follows the        */
/* program. You are also allowed to modify the program for your        */
/* own calculations if you give a reference to the author.             */
/* See also notice below !                                             */
/*                                                                     */
/*                                                                     */
/* IMPORTANT !                                                         */
/* -----------                                                         */
/*                                                                     */
/* Please don't use the result from this program for important         */
/* purposes without doing an independent check of the result !         */
/* See also WARNING and notice below !                                 */
/*                                                                     */
/*                                                                     */
/* Function:                                                           */
/* ---------                                                           */
/*                                                                     */
/* This C program calculates pi, the golden section = (sqrt(5)+1)/2,   */
/* e = exp(1) and sqrt(n) with about 2^power decimals.                 */
/*                                                                     */
/* If for example (power of 2) = 10 the calculations are made with     */
/* about 2^10 = 1024 digits.                                           */ 
/*                                                                     */
/* The program can quite easily be modified for other calculations.    */
/*                                                                     */
/*                                                                     */
/* Accuracy:                                                           */
/* ---------                                                           */
/*                                                                     */
/* All printed digits are probably correct. The results for            */
/* pi and e have been verified up to 2^22 digits.                      */
/*                                                                     */
/* OBSERVE ! The code has been somewhat changed since this             */
/* verification was made !                                             */
/*                                                                     */
/*                                                                     */
/* Compilation:                                                        */
/* ------------                                                        */
/*                                                                     */
/* The program works in UNIX and LINUX and I recommend that you        */
/* optimize the program (i.e. use cc -O2 option or similar option)     */
/* when you compile (gcc if you use a GNU compiler).                   */
/*                                                                     */
/*> cc -O2 -o sspi sspi.c -lm                                          */
/*                                                                     */
/* Observe that on some computers (HP) the code will be faster if the  */
/* variables in the FFT are declared as static. Then remove the        */
/* comments around words "static" in the FFT code.                     */
/*                                                                     */
/* The time wraps around after 4295                                    */
/* seconds on some (HP) computers.                                     */ 
/*                                                                     */
/*                                                                     */
/* Output:                                                             */
/* -------                                                             */
/*                                                                     */
/* Output is to the file "sspi.dat". This file name can easily be      */
/* changed.                                                            */
/*                                                                     */
/*                                                                     */
/* Algorithms:                                                         */
/* -----------                                                         */
/*                                                                     */
/* The program uses:                                                   */
/*                                                                     */
/* * FFT for multiplications.                                          */
/*                                                                     */
/* * Newton-Raphson for divisions.                                     */
/*                                                                     */
/* * Newton-Raphson like 4-th order method for square roots.           */
/*                                                                     */
/* * Borweins 4-th order or Gauss-Legendre algorithm for pi.           */
/*                                                                     */
/*                                                                     */
/* WARNING !                                                           */
/* ---------                                                           */
/*                                                                     */
/* If you want to make calculations with much more than 4 million      */
/* (2^22) digits the convolutions in the FFT multiplications can be    */
/* saturated which will give erroneous results or no convergence       */
/* at all. If this occur then set mulversion=2 which                   */
/* is 2 times slower than mulversion=1 but use the base 100            */
/* instead of 10000. This means that the convolution will be           */
/* saturated for a much larger number of digits (> 100 millions).      */
/* You find the variable mulversion a few lines below this text.       */
/*                                                                     */
/*                                                                     */
/* Memory:                                                             */
/* -------                                                             */
/*                                                                     */
/* The program requires 13.5 Mbyte memory per million digits of        */
/* pi or e if mulversion=1 and 22 Mbyte if mulversion=2.               */
/*                                                                     */
/*                                                                     */
/* Timing:                                                             */
/* -------                                                             */
/*                                                                     */
/* The time complexity for pi is: t = constant*n*log(n)^3,             */
/* Where n is the number of digits.                                    */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Notice:                                                                   */
/* =======                                                                   */
/*                                                                           */
/* I make no warranties that this program is (1) free of error, (2) con-     */
/* sistent with any standard merchantability, or (3) meeting the require-    */
/* ments of a particular application.  This software shall not, partly or    */
/* as a whole, participate in a process, whose outcome can result in injury  */
/* to a person or loss of property. It is solely designed for analytical     */
/* work.  Permission to use, copy, and distribute is hereby granted without  */
/* fee, providing that the header above including this notice appears in     */
/* all copies.                                                               */
/*                                                                           */
/*                                                         Stefan Spaennare  */
/*                                                                           */
/*****************************************************************************/

#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>


double timetic=(double)(CLOCKS_PER_SEC);

char fname[80]="sspienr.dat";  /* Output filename */

int mulversion=1;           /* Multiplication version (1 or 2) */

double bas=10000.0;
double basi=0.0001;

double pi=3.14159265358979323;


void bitrev(d,n)
double d[];
int n;
{
   int i,j,k,m,n2,n22,j2,j22,k2,k22;
   double tr,ti;

   n2=2*n;
   n22=n2-1;
   j=0;
   for (i=0; i<=n22; i++) {
      k=2*i;
      if (j > k) {
         j2=j+1;
         j22=j2+1;
         k2=k+1;
         k22=k2+1;
         tr=d[j2];
         ti=d[j22];
         d[j2]=d[k2];
         d[j22]=d[k22];
         d[k2]=tr;
         d[k22]=ti;
      } /* if */
      m=n2/2;
      while ((m >= 2) && (j >= m)) {
         j=j-m;
         m=m/2;
      } /* while */
      j=j+m;
   } /* for i */

} /* bitrev */


void fft(d,n,ss)
double d[];
int n,ss;
{
   int i,is,j,m,mmax,n2,i2,i22,j2,j22;
   double tr,ti,wr,wpr,wi,wpi,wt,theta,aux,dss,temp;
   double d1r,d1i,d2r,d2i;

   bitrev(d,n);

   if (ss == -1) {
      dss=-1.0;
   }
   else {
      dss=1.0;
   }

   n2=2*n;
   mmax=2;

   while (n2 > mmax) {
      is=2*mmax;
      theta=dss*2.0*pi/(double)(mmax);
      temp=sin(0.5*theta);
      wpr=-2.0*temp*temp;
      wpi=sin(theta);
      wr=1.0;
      wi=0.0;
      m=0;
      while (m < mmax) {
         i=m;
         while (i < n2) {
            j=i+mmax; 
            j2=j+1;
            j22=j2+1;
            i2=i+1;
            i22=i2+1;
            d1r=d[j2];
            d1i=d[j22];
            d2r=d[i2];
            d2i=d[i22];
            tr=wr*d1r-wi*d1i;
            ti=wr*d1i+wi*d1r;
            d[j2]=d2r-tr;
            d[j22]=d2i-ti;
            d[i2]=d2r+tr;
            d[i22]=d2i+ti;
            i=i+is;
         } /* while */
         wt=wr;
         wr=wr*wpr-wi*wpi+wr;
         wi=wi*wpr+wt*wpi+wi;
         m=m+2;
      } /* while */
      mmax=is;
   } /* while */

} /* fft */     


void mul(d,n)

double *d;
int n;

{
   int nn,nn3,nn2,i,j,le,ndim;
   double x1,x2,y1,y2;
   
   nn=n+1;
   nn2=2*n+2;
   nn3=nn2+1;

   x1=d[1]; y1=d[2];
   x2=y1; y1=0.0; y2=0.0;
   d[1]=x1*x2-y1*y2;
   d[2]=x1*y2+y1*x2;

   for (j=3; j<=nn; j=j+2) {
      x1=0.5*(d[j]+d[nn2-j]);
      x2=0.5*(d[j]-d[nn2-j]);
      y1=0.5*(d[j+1]+d[nn3-j]);
      y2=0.5*(d[j+1]-d[nn3-j]);
      
      d[j]=x1*y1-y2*(-x2);
      d[j+1]=x1*(-x2)+y2*y1;
      d[nn2-j]=x1*y1-(-y2)*x2;
      d[nn3-j]=x1*x2+(-y2)*y1;
   } /* for j */

} /* mul */


void lzero(a,nn)

short *a;
int *nn;
 
{
   int i,l,il,dl,ib;

   l=nn[0]; il=nn[1]; dl=nn[2];

   ib=l-il-dl+1;

   for (i=l;i>=ib; i--) {
      a[i]=0;
   } /* for i */
   
   a[l+1]=1;

} /* lzero */

      
void linttol(a,man0,exp0,nn)

short *a;
int man0,exp0;
int *nn;
 
{
   int i,l,il,dl,ib;

   l=nn[0]; il=nn[1]; dl=nn[2];

   ib=l-il-dl+1;

   for (i=l;i>=ib; i--) {
      a[i]=0;
   } /* for i */

   a[l-il+exp0+1]=abs(man0);
  
   if (man0 >= 0) { 
      a[l+1]=1;
   }
   else {
      a[l+1]=-1;
  } /* if */

} /* linttol */


void ppp(t)

int t; 

{
  if (t > 999) {
     printf("%4d",t);
  }
  else if (t > 99) {
     printf("0%3d",t);
  } 
  else if (t > 9) {
     printf("00%2d",t);
  }
  else {
     printf("000%1d",t);
  } /* if */

} /* ppp */


void lprint(a,nn)
 
short *a;
int *nn;

{

   int i,j,k,m,l,il,dl,t;

   l=nn[0]; il=nn[1]; dl=nn[3];

   if (a[l+1]==-1) {
      printf("-");
   } /* if */

   i=l;
   while ((a[i]==0) && (i>(l-il+1))) {
      i--;
   } /* while */

   m=0;

   if (i==(l-il+1)) {

      printf("%4d.",a[i]);

      m++;

      for (j=(l-il); j>=(l-il-dl+1); j--) {
            t=a[j];
            ppp(t);
            m++;
            if ((m % 18)==0) {
              printf("\n");
            } /* if */
      } /* for j */ 
   }
   else {

      printf("%d",a[i]);
      m++;

      for (j=(i-1); j>=(l-il+1); j--) {
            t=a[j];
            ppp(t);
            m++;
            if ((m % 18)==0) {
               printf("\n");
            } /* if */
      } /* for j */ 
     
      printf(".");

      for (j=(l-il); j>=(l-il-dl+1); j--) {
            t=a[j];
            ppp(t);
            m++;
            if ((m % 18)==0) {
               printf("\n");
            } /* if */
      } /* for j */ 
   } /* else */

   printf("\n\n");

} /* lprint */


void fppp(t,outfile)

int t; 
FILE *outfile;

{
  if (t > 999) {
     fprintf(outfile,"%4d",t);
  }
  else if (t > 99) {
     fprintf(outfile,"0%3d",t);
  } 
  else if (t > 9) {
     fprintf(outfile,"00%2d",t);
  }
  else {
     fprintf(outfile,"000%1d",t);
  } /* if */

} /* fppp */


void lfprint(a,nn,outfile)
 
short *a;
int *nn;
FILE *outfile;

{

   int i,j,k,m,l,il,dl,t;

   l=nn[0]; il=nn[1]; dl=nn[3];

   if (a[l+1]==-1) {
      fprintf(outfile,"-");
   } /* if */

   i=l;
   while ((a[i]==0) && (i>(l-il+1))) {
      i--;
   } /* while */

   m=0;

   if (i==(l-il+1)) {

      fprintf(outfile,"%4d.",a[i]);

      m++;

      for (j=(l-il); j>=(l-il-dl+1); j--) {
            t=a[j];
            fppp(t,outfile);
            m++;
            if ((m % 18)==0) {
              fprintf(outfile,"\n");
            } /* if */
      } /* for j */ 
   }
   else {

      fprintf(outfile,"%d",a[i]);
      m++;

      for (j=(i-1); j>=(l-il+1); j--) {
            t=a[j];
            fppp(t,outfile);
            m++;
            if ((m % 18)==0) {
               fprintf(outfile,"\n");
            } /* if */
      } /* for j */ 
     
      fprintf(outfile,".");

      for (j=(l-il); j>=(l-il-dl+1); j--) {
            t=a[j];
            fppp(t,outfile);
            m++;
            if ((m % 18)==0) {
               fprintf(outfile,"\n");
            } /* if */
      } /* for j */ 
   } /* else */

   fprintf(outfile,"\n");

} /* lfprint */



void add(a,b,c,nn)

short *a;
short *b;
short *c;
int *nn;

{
  int i,ib,il,dl,l;

  int q,r,basint;

  l=nn[0]; il=nn[1]; dl=nn[2];

  ib=l-il-dl+1;

  basint=10000;
  r=0;

  for (i=ib; i<=l; i++) {
     q=b[i]+c[i]+r;
     if (q >= basint) {
        a[i]=q-basint;
	r=1;
     }
     else {
        a[i]=q;
	r=0;
     } /* if */
  } /* for i */

} /* add */
      

void sub(a,b,c,nn)

short *a;
short *b;
short *c;
int *nn;

{
  int i,ib,il,dl,l,m,t;

  int q,r,basint;

  l=nn[0]; il=nn[1]; dl=nn[2];

  ib=l-il-dl+1;

  basint=10000;
  r=0;

  for (i=ib; i<=l; i++) {
     q=b[i]-c[i]-r;
     if (q < 0) {
        a[i]=basint+q;
	r=1;
     }
     else {
        a[i]=q;
	r=0;
     } /* if */
  } /* for i */

} /* sub */


int labscomp(a,b,nn)

short *a;
short *b;
int *nn;

{
  int i,l,il,dl,ib;

  l=nn[0]; il=nn[1]; dl=nn[2];

  ib=l-il-dl+1;
  i=l;

  while ((a[i]==b[i]) && (i>=ib)) {
     i--;
  } /* while */

  if (a[i]>=b[i]) {
     return(1);
  }
  else {
     return(-1);
  } /* if */

} /* lasbcomp */


void ladd(a,b,c,nn)

short *a;
short *b;
short *c;
int *nn;

{
   int b2,c2,co,l;

   l=nn[0];

   co=labscomp(b,c,nn);
  
   b2=b[l+1];
   c2=c[l+1]; 

      if ((b2>0) && (c2>0)) {
         add(a,b,c,nn);
         a[l+1]=1;
      } /* if */
      if ((b2>0) && (c2<0) && (co==1)) {
         sub(a,b,c,nn);
         a[l+1]=1;
      } /* if */
      if ((b2>0) && (c2<0) && (co==-1)) {
         sub(a,c,b,nn);
         a[l+1]=-1;
      } /* if */
      if ((b2<0) && (c2>0) && (co==1)) {
         sub(a,b,c,nn);
         a[l+1]=-1;
      } /* if */
      if ((b2<0) && (c2>0) && (co==-1)) {
         sub(a,c,b,nn);
         a[l+1]=1;
      } /* if */
      if ((b2<0) && (c2<0)) {
         add(a,b,c,nn);
         a[l+1]=1;
      } /* if */

} /* ladd */


void lsub(a,b,c,nn)

short *a;
short *b;
short *c;
int *nn;

{
   int b2,c2,co,l;

   l=nn[0];

   co=labscomp(b,c,nn);
  
   b2=b[l+1];
   c2=c[l+1]; 

      if ((b2>0) && (c2>0) && (co==1)) {
         sub(a,b,c,nn);
         a[l+1]=1;
      } /* if */
      if ((b2>0) && (c2>0) && (co==-1)) {
         sub(a,c,b,nn);
         a[l+1]=-1;
      } /* if */
      if ((b2>0) && (c2<0)) {
         add(a,b,c,nn);
         a[l+1]=1;
      } /* if */
      if ((b2<0) && (c2>0)) {
         add(a,b,c,nn);
         a[l+1]=-1;
      } /* if */
      if ((b2<0) && (c2<0) && (co==1)) {
         sub(a,c,b,nn);
         a[l+1]=-1;
      } /* if */
      if ((b2<0) && (c2<0) && (co==-1)) {
         sub(a,b,c,nn);
         a[l+1]=1;
      } /* if */

} /* lsub */


void lcmul(a,b,c,nn,exp0)

short *a;
int b;
short *c;
int *nn;
int exp0;

{
   int i,j,k,l,il,dl,ib,b2;
   double q,r,b2r;

   l=nn[0]; il=nn[1]; dl=nn[2];

   if (b>=0) {
      b2=b;
   }
   else {
      b2=-b;
   } /* if */

   ib=l-il-dl+1;
   k=ib+exp0;

   for (i=ib; i<k; i++) {
      a[i]=0;
   } /* for i */

   r=0.0;
   b2r=(double)(b2);

   for (j=ib; j<=l; j++) {
      q=(double)(c[j])*b2r+r+0.5;
      r=(double)((int)(basi*q));
      if (k>=ib) {
         a[k]=(int)(q-bas*r);
      } /* if */
      k++;
   } /* for j */

   a[k]=(int)(r+0.5);

   k++;
   for (i=k; i<=l; i++) {
      a[i]=0;
   } /* for i */

   if (b>=0) {
      a[l+1]=c[l+1];
   }
   else {
      a[l+1]=-c[l+1];
   } /* if */

} /* lcmul */


void lcmul2(a,b,c,nn,exp0)

short *a;
double b;
short *c;
int *nn;
int exp0;

{
   int i,j,k,l,il,dl,ib;
   double q,r,b2r,b2;

   l=nn[0]; il=nn[1]; dl=nn[2];

   if (b>=0.0) {
      b2=b;
   }
   else {
      b2=-b;
   } /* if */

   ib=l-il-dl+1;
   k=ib+exp0;

   for (i=ib; i<k; i++) {
      a[i]=0;
   } /* for i */

   r=0.0;
   b2r=b2;

   for (j=ib; j<=l; j++) {
      q=(double)(c[j])*b2r+r+0.5;
      r=floor(basi*q);
      if (k>=ib) {
         a[k]=(int)(q-bas*r);
      } /* if */
      k++;
   } /* for j */

   a[k]=(int)(r+0.5);

   k++;
   for (i=k; i<=l; i++) {
      a[i]=0;
   } /* for i */

   if (b>=0.0) {
      a[l+1]=c[l+1];
   }
   else {
      a[l+1]=-c[l+1];
   } /* if */

} /* lcmul2 */


void ltoreal(a,man0,exp0,nn)

short *a;
double *man0,*exp0;
int *nn;

{
   char s0[20];
   char s1[10];
   char s2[10];
   char s3[10];
   char s4[10];

   char st1[10];
   char st2[10];
   char st3[10];
   char st4[10];

   int i,n,l,il,dl,ib,length,le;

   double man,ex;

   l=nn[0]; il=nn[1]; dl=nn[2];

   ib=l-il-dl+1;

   i=l;

   while ((a[i]==0) && (i>=ib)) {
      i--;
   } /* while */

   if ((i==ib) && (a[i]==0)) {
      man=0.0;
      ex=0.0;
   }
   else {
      sprintf(s1,"%d",a[i]);
      length=strlen(s1);
      ex=(double)(4*(i-(l-il+1))+length);

      sprintf(s2,"%d",a[i-1]);
      sprintf(s3,"%d",a[i-2]);
      sprintf(s4,"%d",a[i-3]);
 
      strcpy(st2,"");
      le=strlen(s2);
      for (i=1; i<=(4-le); i++) {
         strcat(st2,"0");
      } /* for i */
      strcat(st2,s2);

      strcpy(st3,"");
      le=strlen(s3);
      for (i=1; i<=(4-le); i++) {
         strcat(st3,"0");
      } /* for i */
      strcat(st3,s3);

      strcpy(st4,"");
      le=strlen(s4);
      for (i=1; i<=(4-le);i++) {
         strcat(st4,"0");
      } /* for i */
      strcat(st4,s4);

      sprintf(s0,"%s%s%s%s%s","0.",s1,st2,st3,st4);
      man=atof(s0);
   } /* if */
   
   *man0=(double)(a[l+1])*man;
   *exp0=ex;

} /* ltoreal */


void lrealtol(a,man0,exp0,nn)

short *a;
double man0,exp0;
int *nn;

{
    char s0[18];
    char s1[16];
 
    char z1[7];
    char z2[7];
    char z3[7];
    char z4[7];

    double man,ex,exp2,man2;

    int i,k,n,l,il,dl,ib,l1,l2,sl,ll;

    l=nn[0]; il=nn[1]; dl=nn[2];

    ib=l-il-dl+1;

    for (i=l; i>=ib; i--) {
       a[i]=0;
    } /* for i */   

    man2=man0;
    ex=exp0;
    man=fabs(man2);

    while (man>=1.0) {
       man=man/10.0;
       ex=ex+1.0;
    } /* while */

    sprintf(s0,"%17.16f",man);
 
    for (i=2; i<=17; i++) {
       s1[i-2]=s0[i];
    } /* for i */


    exp2=fabs(ex/(double)(4));
    l1=(int)(ex/(double)(4));
    l2=(int)(4*(exp2-(double)(abs(l1))));
    
    if (ex<0) {
       l2=4-l2;
       l1=l1-1;
    } /* if */

    if (l2==0) {
       l2=4;
       l1=l1-1; 
    } /* if */

    for (i=0; i<=l2-1; i++) {
       z1[i]=s1[i];
    } /* for i */

    z1[i]='\0';

    for (i=0; i<=4-1; i++) {
       z2[i]=s1[l2+i];
    } /* for i */

    z2[i]='\0';

    k=0;

    for (i=0; i<=4-1; i++) {
       if ((l2+i+4)<=16) {
          z3[i]=s1[l2+i+4];
          k++;
       } /* if */
    } /* for i */
 
    z3[k]='\0'; 

    for (i=1; i<=4-k; i++) {
       strcat(z3,"0");
    } /* for i */

    k=0;

    for (i=0; i<=4-1; i++) {
       if ((l2+i+2*4)<=16) {
          z4[i]=s1[l2+i+2*4];
          k++;
       } /* if */
    } /* for i */
 
    z4[k]='\0'; 

    for (i=1; i<=4-k; i++) {
       strcat(z4,"0");
    } /* for i */
    
    ll=l-il+l1+1;
 
    a[ll]=atoi(z1);
    a[ll-1]=atoi(z2);
    a[ll-2]=atoi(z3);
    a[ll-3]=atoi(z4);

    if (man2>=0) {
       a[l+1]=1;
    }
    else {
       a[l+1]=-1;
    } /* if */

} /* lrealtol */


int lcompeps(a,nn)

short *a;
int *nn;

{
   int i,l,il,dl,k,ib,n,ll;

   l=nn[0]; il=nn[1]; dl=nn[2]; n=nn[3];

   ib=l-il-dl+1;

   k=0;

   ll=l-il-n;

   for (i=l; i>=ll; i--) {
      if (a[i]==0) {
         k++;
      } /* if */
   } /* for i */

   if (k==(il+n+1)) {
      return(-1);
   }
   else {
      return(1);
   } /* if */

}  /* lcompeps */


int lgetdec(a,nn)

short *a;
int *nn;

{
   int i,l,il,dl,k,ib;

   l=nn[0]; il=nn[1]; dl=nn[2];

   ib=l-il-dl+1;

   k=0;

   i=l-il;

   while ((a[i]==0) && (i>ib)) {
      k++;
      i--;
   } /* while */

   return(4*k);

}  /* lgetdec */


void lass(a,b,nn)

short *a;
short *b;
int *nn;

{
  int i,l,il,dl,ib;

  l=nn[0]; il=nn[1]; dl=nn[2];
  
  ib=l-il-dl+1;

  for (i=l+1; i>=ib; i--) {
     a[i]=b[i];
  } /* for i */

} /* lass */


void lmul(a,b,c,nn,q)
short *a;
short *b;
short *c;
int *nn;
double *q;

{


   int i,j,k,l,il,dl,ib,n,n2;
   int bi,b1,c1;

   double isize,rr,qq,xx,yy,x1,x2,y1,y2;
   double br,bri;


if (mulversion == 1) {

   l=nn[0]; il=nn[1]; dl=nn[2];

   n=2*l;
   n2=2*n;

   isize=1.0/sqrt((double)(n));

   ib=l-il-dl+1;
   
   for (i=1; i<=n2; i++) {
      q[i]=0.0;
   } /* for i */
 
   k=1;

   for (i=ib; i<=l; i++) {
      q[k]=(double)(b[i])*isize;
      q[k+1]=(double)(c[i])*isize;
      k=k+2;
   } /* for i */


   fft(q,n,-1);

   mul(q,n);

   fft(q,n,1);

   rr=0.0;

   for (i=1; i<=n2; i=i+2) {
      xx=q[i]+rr;
      rr=floor((xx+0.5)*basi);
      q[i]=xx-bas*rr; 
   } /* for i */

   k=n-2*il+1;

   for (i=ib; i<=l; i++) {
      a[i]=(int)(q[k]+0.5);
      k=k+2;
   } /* for i */

   a[l+1]=b[l+1]*c[l+1]; 

}
else {
  
   l=nn[0]; il=nn[1]; dl=nn[2];

   n=4*nn[0];
   n2=2*n;

   br=100.0;
   bri=0.01;
   bi=100;

   isize=1.0/sqrt((double)(n));

   ib=l-il-dl+1;
   
   for (i=1; i<=n2; i++) {
      q[i]=0.0;
   } /* for i */
 
   k=1;

   for (i=ib; i<=l; i++) {
      b1=b[i] % bi;
      q[k]=(double)(b1)*isize;
      b1=b[i] / bi;
      q[k+2]=(double)(b1)*isize;
      c1=c[i] % bi;
      q[k+1]=(double)(c1)*isize;
      c1=c[i] / bi;
      q[k+3]=(double)(c1)*isize;
      k=k+4;
   } /* for i */

   fft(q,n,-1);
 
   mul(q,n);

   fft(q,n,1);

   rr=0.0;

   for (i=1; i<=n2; i=i+2) {
      xx=q[i]+rr;
      rr=floor((xx+0.5)*bri);
      q[i]=xx-br*rr; 
   } /* for i */

   k=n-4*il+1;

   for (i=ib; i<=l; i++) {
      a[i]=bi*(int)(q[k+2]+0.5)+(int)(q[k]+0.5);
      k=k+4;
   } /* for i */

   a[l+1]=b[l+1]*c[l+1]; 

} /* if */


} /* lmul */


 
void linv(a,b,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   double man,ex;
   
   int s,l,il,dl;
 
   l=nn[0]; il=nn[1]; dl=nn[2];
 
   s=b[l+1];
   b[l+1]=1;

   ltoreal(b,&man,&ex,nn);

   man=1.0/man;
   ex=-ex;

   lrealtol(x1,man,ex,nn);

   do {
      lass(x2,x1,nn);
      lcmul(a,2,x1,nn,0);
      lmul(t1,b,x1,nn,q);
      lmul(t2,x1,t1,nn,q);
      lsub(x1,a,t2,nn);
      lsub(t1,x1,x2,nn);
   } while (lcompeps(t1,nn)==1); 

   lass(a,x1,nn);

   a[l+1]=s;

} /* linv */ 
 
 
void ldivv(a,b,c,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *c;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   linv(a,c,x1,x2,t1,t2,nn,q);

   lmul(a,a,b,nn,q);

} /* ldivv */    

 
void lsqrtinv(a,b,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   double re0,re1,re2,s10,man,ex,eps;

   int s,l,il,dl;
 
   l=nn[0]; il=nn[1]; dl=nn[2];

   eps=0.001; 
   s=b[l+1];
   b[l+1]=1;

   ltoreal(b,&man,&ex,nn);

   re0=1.0/sqrt(man);
   re1=-ex/2.0;
   re2=fabs(re1);

   s10=sqrt(10.0);

   if ((re2-(double)((int)(re2)))>eps) {
      re0=re0/s10;
      re1=re1+0.5;
   } /* if */

   man=re0;
   ex=re1;

   lrealtol(x1,man,ex,nn);

   do {
      lass(x2,x1,nn);
      lmul(t1,x1,b,nn,q);
      lmul(a,x1,t1,nn,q);
      lcmul(t1,5,a,nn,0);
      linttol(x1,21,0,nn);
      lsub(t2,x1,t1,nn);
      lmul(t1,a,t2,nn,q);
      linttol(x1,-35,0,nn);
      ladd(t2,t1,x1,nn);
      lmul(t1,a,t2,nn,q);
      linttol(x1,35,0,nn);
      ladd(t2,t1,x1,nn);
      lmul(t1,x2,t2,nn,q);
      lcmul(x1,625,t1,nn,-1);
      lsub(t1,x1,x2,nn);
   } while (lcompeps(t1,nn)==1); 

   lass(a,x1,nn);

} /* lsqrtinv */


void lsqrt(a,b,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   lsqrtinv(a,b,x1,x2,t1,t2,nn,q);

   lmul(a,a,b,nn,q);

} /* lsqrt */


void lssqrtinv(a,b,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   double re0,re1,re2,s10,man,ex,eps;

   int s,l,il,dl;
 
   l=nn[0]; il=nn[1]; dl=nn[2];
 
   eps=0.001;
   s=b[l+1];
   b[l+1]=1;

   ltoreal(b,&man,&ex,nn);

   re0=1.0/sqrt(sqrt(man));
   re1=-ex/4.0;
   re2=fabs(re1);

   s10=sqrt(sqrt(10.0));

   while ((re2-(double)((int)(re2)))>eps) {
      re0=re0/s10;
      re1=re1+0.25;
      re2=fabs(re1);
   } /* while */

   man=re0;
   ex=re1;

   lrealtol(x1,man,ex,nn);

   do {
      lass(x2,x1,nn);
      lmul(t1,x1,x1,nn,q);
      lmul(t1,t1,t1,nn,q);
      lmul(a,t1,b,nn,q);
      lcmul(t1,15,a,nn,0);
      linttol(x1,65,0,nn);
      lsub(t2,x1,t1,nn);
      lmul(t1,a,t2,nn,q);
      linttol(x1,-117,0,nn);
      ladd(t2,t1,x1,nn);
      lmul(t1,a,t2,nn,q);
      linttol(x1,195,0,nn);
      ladd(t2,t1,x1,nn);
      lmul(t1,x2,t2,nn,q);
      lcmul(x1,781250,t1,nn,-2);
      lsub(t1,x1,x2,nn);
   } while (lcompeps(t1,nn)==1); 

   lass(a,x1,nn);

} /* lssqrtinv */


void lssqrt(a,b,x1,x2,t1,t2,nn,q)

short *a;
short *b;
short *x1;
short *x2;
short *t1;
short *t2;
int *nn;
double *q;

{

   lssqrtinv(a,b,x1,x2,t1,t2,nn,q);

   lmul(t1,a,a,nn,q);
   lmul(t2,t1,a,nn,q);
   lmul(a,t2,b,nn,q);

} /* lssqrt */


int main(argc,argv)
int argc;
char *argv[];

{

   FILE *outfile;

   int ff,il,ib,dl,ll2,ll;
   int i,j,k,length,ilength,dec,pot,l,nr,n;
   int f2,iel,m,ql;
   int qqi,remi,basint,l2,l22;

   double frac,el,le,dll,lt,r,q2;
   double r2,c2,cl,jd,bbas,bbasi,iif;

   int nn[4];

   clock_t time1,timediff1,time2,timediff2,time3,timediff3;

   short *t1,*t2,*x1,*x2,*a,*b,*c,*d,*e,*f; 

   int *aa,*bb;

   double *q;


   printf("\n\n");
   printf("                 O\n");                      
   printf("    OOOOOOOOOOOOO\n");
   printf("   OOOOOOOOOOOOO\n"); 
   printf("  O  OO     OO\n");
   printf("     OO     OO\n");        
   printf("     OO     OO\n");
   printf("     OO     OO\n");
   printf("    OO       OO\n");
   printf("   OO         OO\n");
   printf("\n\n");
   printf("0: Quit\n\n");
   printf("1: Pi, Borweins 4-th order (fastest)\n\n");
   printf("2: Pi, Gauss-Legendre\n\n");
   printf("3: The Golden Section = (sqrt(5)+1)/2\n\n");
   printf("4: Sqrt(n) (n integer)\n\n");
   printf("5: e = exp(1)\n\n");
   printf("Input nr: ");

   scanf("%d",&nr);

   printf("\n\n");


   ql=4*mulversion;
 


if (nr == 1) {


   printf("Pi, Borweins 4-th order\n");
   printf("=======================\n\n");

   printf("Input power of 2:  ");

   scanf("%d",&pot);


   printf("\n\n");
  
   pot=pot-2;

   length=1;

   for (i=1; i<=pot; i++) {
      length=2*length;
   } /* for i */

   ilength=2;
   frac=0.995;

   nn[0]=length;
   nn[1]=ilength;
   nn[2]=length-ilength;
   nn[3]=(int)(frac*(double)(length-ilength))-2;

   l=length;

   t1 = (short *)calloc(l+2,sizeof(short));
   t2 = (short *)calloc(l+2,sizeof(short));
   x1 = (short *)calloc(l+2,sizeof(short));
   x2 = (short *)calloc(l+2,sizeof(short));
   a = (short *)calloc(l+2,sizeof(short));
   b = (short *)calloc(l+2,sizeof(short));
   c = (short *)calloc(l+2,sizeof(short));
   d = (short *)calloc(l+2,sizeof(short));
   e = (short *)calloc(l+2,sizeof(short));
   f = (short *)calloc(l+2,sizeof(short));

   q = (double *)calloc(ql*l+1,sizeof(double));


   time2=clock();


   printf("Decimals: %10d     FFT: %10d\n",4*nn[3],ql*2*length/4);

   printf("\n\nCalculating ...\n");

   lzero(a,nn);
   lzero(b,nn);
   lzero(c,nn);
   lzero(d,nn);
   lzero(e,nn);
   lzero(f,nn);

   linttol(d,2,0,nn);

   time3=clock(); 

   lsqrt(c,d,x1,x2,t1,t2,nn,q);

   timediff3=clock()-time3; 
   printf("\nTime sqrt:      %10.2f\n",(double)(timediff3)/timetic);

   linttol(d,1,0,nn);

   lsub(b,c,d,nn);


   linttol(e,6,0,nn);

   lcmul(d,4,c,nn,0);

   lsub(a,e,d,nn); 


   printf("\n    it      decimals       scpu\n\n");

   time1=clock();

   k=1;
   iif=8.0;

   do {
      time3=clock();

      lmul(c,b,b,nn,q);
      lmul(c,c,c,nn,q);
      linttol(d,1,0,nn);
      lsub(e,d,c,nn);

      lssqrt(b,e,x1,x2,t1,t2,nn,q);
      lsub(e,d,b,nn);
      ladd(c,d,b,nn);
      ldivv(b,e,c,x1,x2,t1,t2,nn,q);

      ladd(c,b,d,nn);
      lmul(c,c,c,nn,q);
      lmul(c,c,c,nn,q);
      lmul(c,a,c,nn,q);
      
      lmul(e,b,b,nn,q);
      ladd(f,b,e,nn);
      ladd(e,f,d,nn);
      lmul(f,e,b,nn,q);

      lcmul2(d,iif,f,nn,0);
      lsub(a,c,d,nn); 
      dec=lgetdec(d,nn);
      timediff3=clock()-time3;
      printf("%6d  %12d %10.2f\n",k,dec,(double)(timediff3)/timetic);
      iif=4.0*iif;
      k++;
   } while (lcompeps(d,nn)==1); 


   timediff1=clock()-time1;

   printf("\nTime main loop: %10.2f\n",(double)(timediff1)/timetic);


   time3=clock();

   linv(c,a,x1,x2,t1,t2,nn,q);
 
   timediff3=clock()-time3;
   printf("\nTime div:       %10.2f\n",(double)(timediff3)/timetic);

   timediff2=clock()-time2;
   printf("\nTime total:     %10.2f\n",(double)(timediff2)/timetic);


   printf("\n");

   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   lfprint(c,nn,outfile);  

   printf("\n");


} /* if */



if (nr == 2) {


   printf("Pi, Gauss-Legendre\n");
   printf("==================\n\n");

   printf("Input power of 2:  ");

   scanf("%d",&pot);

   printf("\n\n");

   pot=pot-2;

   length=1;

   for (i=1; i<=pot; i++) {
      length=2*length;
   } /* for i */

   ilength=2;
   frac=0.995;

   nn[0]=length;
   nn[1]=ilength;
   nn[2]=length-ilength;
   nn[3]=(int)(frac*(double)(length-ilength))-2;

   l=length;

   t1 = (short *)calloc(l+2,sizeof(short));
   t2 = (short *)calloc(l+2,sizeof(short));
   x1 = (short *)calloc(l+2,sizeof(short));
   x2 = (short *)calloc(l+2,sizeof(short));
   a = (short *)calloc(l+2,sizeof(short));
   b = (short *)calloc(l+2,sizeof(short));
   c = (short *)calloc(l+2,sizeof(short));
   d = (short *)calloc(l+2,sizeof(short));
   e = (short *)calloc(l+2,sizeof(short));

   q = (double *)calloc(ql*l+1,sizeof(double));


   time2=clock();


   printf("Decimals: %10d     FFT: %10d\n",4*nn[3],ql*2*length/4);

   printf("\n\nCalculating ...\n");

   lzero(a,nn);
   lzero(b,nn);
   lzero(c,nn);
   lzero(d,nn);
   lzero(e,nn);

   linttol(a,1,0,nn);

   linttol(d,2,0,nn);

   time3=clock(); 

   lsqrtinv(b,d,x1,x2,t1,t2,nn,q);


   timediff3=clock()-time3; 
   printf("\nTime sqrt:      %10.2f\n",(double)(timediff3)/timetic);


   printf("\n    it      decimals       scpu\n\n");

   time1=clock();

   k=1;
   iif=4.0;

   do {
      time3=clock();
      lmul(c,a,b,nn,q);
      ladd(t1,a,b,nn);
      lcmul(a,5000,t1,nn,-1);
      lsqrt(b,c,x1,x2,t1,t2,nn,q);
      lmul(t2,a,a,nn,q);
      lsub(d,t2,c,nn);
      lcmul2(t2,iif,d,nn,0);
      ladd(t1,e,t2,nn);
      lass(e,t1,nn);
      dec=lgetdec(d,nn);
      timediff3=clock()-time3;
      printf("%6d  %12d %10.2f\n",k,dec,(double)(timediff3)/timetic);
      iif=2.0*iif;
      k++;
   } while (lcompeps(d,nn)==1); 


   timediff1=clock()-time1;

   printf("\nTime main loop: %10.2f\n",(double)(timediff1)/timetic);

   ladd(t1,a,b,nn);

   lass(t2,t1,nn);

   lmul(a,t1,t2,nn,q);

   linttol(t1,1,0,nn);

   lsub(b,t1,e,nn);

   time3=clock();

   ldivv(c,a,b,x1,x2,t1,t2,nn,q);
 
   timediff3=clock()-time3;
   printf("\nTime div:       %10.2f\n",(double)(timediff3)/timetic);

   timediff2=clock()-time2;
   printf("\nTime total:     %10.2f\n",(double)(timediff2)/timetic);


   printf("\n");

   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   lfprint(c,nn,outfile);  

   printf("\n");

} /* if */



if (nr == 3) {


   printf("The Golden Section = (sqrt(5)+1)/2\n");
   printf("==================================\n\n");

   printf("Input power of 2:  ");

   scanf("%d",&pot);

   printf("\n\n");

   pot=pot-2;

   length=1;

   for (i=1; i<=pot; i++) {
      length=2*length;
   } /* for i */

   ilength=2;
   frac=0.995;

   nn[0]=length;
   nn[1]=ilength;
   nn[2]=length-ilength;
   nn[3]=(int)(frac*(double)(length-ilength))-2;

   l=length;

   t1 = (short *)calloc(l+2,sizeof(short));
   t2 = (short *)calloc(l+2,sizeof(short));
   x1 = (short *)calloc(l+2,sizeof(short));
   x2 = (short *)calloc(l+2,sizeof(short));
   a = (short *)calloc(l+2,sizeof(short));
   b = (short *)calloc(l+2,sizeof(short));
   c = (short *)calloc(l+2,sizeof(short));

   q = (double *)calloc(ql*l+1,sizeof(double));


   time2=clock();


   printf("Decimals: %10d     FFT: %10d\n",4*nn[3],ql*2*length/4);

   printf("\n\nCalculating ...\n");

   lzero(a,nn);
   lzero(b,nn);
   lzero(c,nn);

   time3=clock(); 

   linttol(a,5,0,nn);

   lsqrt(b,a,x1,x2,t1,t2,nn,q);

   linttol(c,1,0,nn);

   ladd(a,b,c,nn);

   lcmul(b,5000,a,nn,-1);

   timediff3=clock()-time3; 
   printf("\nTime GS:      %10.2f\n",(double)(timediff3)/timetic);

   printf("\n");

   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   lfprint(b,nn,outfile);  

   printf("\n");


} /* if */


if (nr == 4) {


   printf("Sqrt(n), n integer\n");
   printf("==================\n\n");

   printf("Input n (1 <= n <= 9999): ");

   scanf("%d",&n);

   if ((n < 1) || (n > 9999)) {
      exit(0);
   } /* if */


   printf("\n");

   printf("Input power of 2:  ");

   scanf("%d",&pot);

   printf("\n\n");

   pot=pot-2;

   length=1;

   for (i=1; i<=pot; i++) {
      length=2*length;
   } /* for i */

   ilength=2;
   frac=0.995;

   nn[0]=length;
   nn[1]=ilength;
   nn[2]=length-ilength;
   nn[3]=(int)(frac*(double)(length-ilength))-2;

   l=length;

   t1 = (short *)calloc(l+2,sizeof(short));
   t2 = (short *)calloc(l+2,sizeof(short));
   x1 = (short *)calloc(l+2,sizeof(short));
   x2 = (short *)calloc(l+2,sizeof(short));
   a = (short *)calloc(l+2,sizeof(short));
   b = (short *)calloc(l+2,sizeof(short));

   q = (double *)calloc(ql*l+1,sizeof(double));


   time2=clock();


   printf("Decimals: %10d     FFT: %10d\n",4*nn[3],ql*2*length/4);

   printf("\n\nCalculating ...\n");

   lzero(a,nn);
   lzero(b,nn);

   time3=clock(); 

   linttol(a,n,0,nn);

   lsqrt(b,a,x1,x2,t1,t2,nn,q);

   timediff3=clock()-time3; 
   printf("\nTime Sqrt:      %10.2f\n",(double)(timediff3)/timetic);


   printf("\n");

   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   lfprint(b,nn,outfile);  

   printf("\n");


} /* if */


if (nr == 5) {


   printf("e = exp(1)\n");
   printf("==========\n\n");

   printf("Input power of 2:  ");

   scanf("%d",&pot);

   printf("\n\n");

   pot=pot-2;

   length=1;

   for (i=1; i<=pot; i++) {
      length=2*length;
   } /* for i */

   ilength=2;

   nn[0]=length;
   nn[1]=ilength;
   nn[2]=length-ilength;
   nn[3]=(int)(0.995*(double)(length-ilength))-2;

   l=length;
   l2=l/2;
   l22=l2+4;

   t1 = (short *)calloc(l+2,sizeof(short));
   t2 = (short *)calloc(l+2,sizeof(short));
   x1 = (short *)calloc(l+2,sizeof(short));
   x2 = (short *)calloc(l+2,sizeof(short));
   a = (short *)calloc(l+2,sizeof(short));
   b = (short *)calloc(l+2,sizeof(short));
   c = (short *)calloc(l+2,sizeof(short));

   aa = (int *)calloc(l22+1,sizeof(int));
   bb = (int *)calloc(l22+1,sizeof(int));

   q = (double *)calloc(ql*l+1,sizeof(double));
  

   printf("Decimals: %10d     FFT: %10d\n\n\n",4*nn[3],ql*2*length/4);

   time2=clock();

   lzero(a,nn);
   lzero(b,nn);
   lzero(c,nn);

   for (i=1; i<=l22; i++) {
      aa[i]=0;
      bb[i]=0;
   } /* for i */

   bbas=100000000.0;
   bbasi=0.00000001;

   basint=100000000;

   lt=1.0/log(10.0);

   el=0.0;

   le=(double)(4*nn[2]);

   i=1;

   while (el<le) {
      el=el+lt*log((double)(i));
      i++;
   } /* while */

   n=i-1;

   ff=(int)(el);

   printf("Factorial=%8d   Length=%8d\n",n,ff); 

   printf("\n\nCalculating ...\n");

   l=nn[0];
   il=nn[1];
   dl=nn[2];
   ib=l-il-dl+1;

   aa[1]=1;

   dll=0.0;

   time1=clock();
   
   for (j=n; j>=1; j--) {
      jd=(double)(j);
      dll=dll+lt*log(jd);
      ll=(int)(0.125*dll+0.5)+3;

      r=0.0;
      for (i=1; i<=ll; i++) {
         q2=(double)(aa[i])*jd+r+0.5;
         r=(double)((int)(bbasi*q2));
         aa[i]=(int)(q2-bbas*r);
      } /* for i */

      remi=0;
      for (i=1; i<=ll; i++) {
         qqi=bb[i]+aa[i]+remi;
	 if (qqi >= basint) {
	    bb[i]=qqi-basint;
	    remi=1;
	 }
	 else {
	    bb[i]=qqi;
	    remi=0;
	 } /* if */
      } /* for i */
   } /* for j */


   timediff1=clock()-time1;
   printf("\nTime main loop: %10.2f\n",(double)(timediff1)/timetic);

   ll2=ib+(int)(0.25*dll+0.5)+2;
   k=1;

   for (i=ib; i<=ll2; i=i+2) {
      a[i]=aa[k] % 10000;
      a[i+1]=aa[k] / 10000;
      b[i]=bb[k] % 10000;
      b[i+1]=bb[k] / 10000;
      k++;
   } /* for i */

   a[l+1]=1;
   b[l+1]=1;

   time3=clock();

   ldivv(c,b,a,x1,x2,t1,t2,nn,q);

   timediff3=clock()-time3;
   printf("\nTime div:       %10.2f\n",(double)(timediff3)/timetic);

   timediff2=clock()-time2;
   printf("\nTime total e = exp(1): %10.2f\n",(double)(timediff2)/timetic);

   printf("\n");

   printf("Calculation finished, output to %s.\n\n",fname);

   outfile=fopen(fname,"w");

   lfprint(c,nn,outfile);  

   printf("\n");

} /* if */


} /* End */
